/*

*/

/**
   This is an example of an unsynchronized
   producer/consumer pair.

   The "bounded buffer" shared by the producer
   and consumer holds only one item.

   The one item in the shared buffer is a pair
   of numbers and the two items in the pair
   should be updated atomically.

   In this example, there is no synchronization
   that prevents the producer from "overflowing",
   or the consumer from "underflowing", their
   shared "bounded buffer".

   In addition, there is no synchronization that
   forces the producer and consumer to update the
   pair of numbers in the buffer atomically.
*/
public class ProducerConsumer
{
   public static void main(String[] args)
   {
      Buffer buf = new Buffer();
      Producer producer = new Producer(buf, 1);
      Consumer consumer = new Consumer(buf, 1);
      Thread t1 = new Thread(producer);
      Thread t2 = new Thread(consumer);
      t1.start();
      t2.start();
      try
      {
         Thread.sleep(10000);

         //t1.interrupt();
         //t2.interrupt();

         t1.join();
         t2.join();
      }
      catch (InterruptedException e)
      {
      }
   }
}


class Buffer
{
   public int counter1 = -1;  // Two data items that should
   public int counter2 = -1;  // be updated atomically.
}


class Producer implements Runnable
{
   private final Buffer buf;
   private final int milliSeconds;

   public Producer(Buffer buf)
   {
      this.buf = buf;
      this.milliSeconds = 500;
   }

   public Producer(Buffer buf, int milliSeconds)
   {
      this.buf = buf;
      this.milliSeconds = milliSeconds;
   }

   public void run()
   {
      int counter1 = 1;
      int counter2 = 10_001;
      try
      {
         while(true)
         {
            buf.counter1 = counter1;
            //Thread.yield(); // force race conditions to happen more often
            buf.counter2 = counter2;
            System.err.println("\n"+counter1+", "+counter2);
            ++counter1;
            ++counter2;
            Thread.sleep(milliSeconds);
         }
      }
      catch (InterruptedException e)
      {
      }
   }
}


class Consumer implements Runnable
{
   private final Buffer buf;
   private final int milliSeconds;

   public Consumer(Buffer buf)
   {
      this.buf = buf;
      this.milliSeconds = 500;
   }

   public Consumer(Buffer buf, int milliSeconds)
   {
      this.buf = buf;
      this.milliSeconds = milliSeconds;
   }

   public void run()
   {
      try
      {
         while(true)
         {
            int counter1 = buf.counter1;
            //Thread.yield(); // force race conditions to happen more often
            int counter2 = buf.counter2;
            if (counter2 - counter1 != 10_000)
            {
               System.out.println("\n\tBINGO: "+counter1+", "+counter2);
               System.exit(1);
            }
            else
            {
               System.out.println("\n\t"+counter1+", "+counter2);
            }
            Thread.sleep(milliSeconds);
         }
      }
      catch (InterruptedException e)
      {
      }
   }
}
